import argparse


def parse_args() -> argparse.Namespace:
    usage = """
        zulip-botserver --config-file <path/to/botserverrc> [--hostname=<address>] [--port=<port>]
    """

    parser = argparse.ArgumentParser(usage=usage)
    mutually_exclusive_args = parser.add_mutually_exclusive_group(required=True)
    # config-file or use-env-vars made mutually exclusive to prevent conflicts
    mutually_exclusive_args.add_argument(
        "--config-file",
        "-c",
        action="store",
        help="Config file for the Botserver. Use your `botserverrc` for multiple bots or"
        "`zuliprc` for a single bot.",
    )
    mutually_exclusive_args.add_argument(
        "--use-env-vars",
        "-e",
        action="store_true",
        help="Load configuration from JSON in ZULIP_BOTSERVER_CONFIG environment variable.",
    )
    parser.add_argument(
        "--bot-config-file",
        action="store",
        default=None,
        help="Config file for bots. Only needed when one of "
        "the bots you want to run requires a config file.",
    )
    parser.add_argument(
        "--bot-name",
        "-b",
        action="store",
        help="Run a single bot BOT_NAME. Use this option to run the Botserver "
        "with a `zuliprc` config file.",
    )
    parser.add_argument(
        "--hostname",
        action="store",
        default="127.0.0.1",
        help="Address on which you want to run the Botserver. (default: %(default)s)",
    )
    parser.add_argument(
        "--port",
        action="store",
        default=5002,
        type=int,
        help="Port on which you want to run the Botserver. (default: %(default)d)",
    )
    return parser.parse_args()



import json
import os
from collections import OrderedDict
from importlib import import_module
from pathlib import Path
from types import ModuleType
from typing import Any, Dict
from unittest import mock

import importlib_metadata as metadata
from typing_extensions import override

from zulip_bots.lib import AbstractBotHandler
from zulip_botserver import server
from zulip_botserver.input_parameters import parse_args

from .server_test_lib import BotServerTestCase


class BotServerTests(BotServerTestCase):
    class MockMessageHandler:
        def handle_message(self, message: Dict[str, str], bot_handler: AbstractBotHandler) -> None:
            assert message == {"key": "test message"}

    class MockLibModule:
        def handler_class(self) -> Any:
            return BotServerTests.MockMessageHandler()

    @override
    def setUp(self) -> None:
        # Since initializing Client invokes `get_server_settings` that fails in the test
        # environment, we need to mock it to pretend that there exists a backend.
        super().setUp()
        self.patch = mock.patch("zulip.Client.get_server_settings", return_value=mock.Mock())
        self.patch.start()

    def test_successful_request(self) -> None:
        available_bots = ["helloworld"]
        bots_config = {
            "helloworld": {
                "email": "helloworld-bot@zulip.com",
                "key": "123456789qwertyuiop",
                "site": "http://localhost",
                "token": "abcd1234",
            }
        }
        self.assert_bot_server_response(
            available_bots=available_bots,
            bots_config=bots_config,
            event=dict(
                message={"content": "@**test** test message"},
                bot_email="helloworld-bot@zulip.com",
                trigger="mention",
                token="abcd1234",  # noqa: S106
            ),
            expected_response="beep boop",
            check_success=True,
        )

    def test_successful_request_from_two_bots(self) -> None:
        available_bots = ["helloworld", "help"]
        bots_config = {
            "helloworld": {
                "email": "helloworld-bot@zulip.com",
                "key": "123456789qwertyuiop",
                "site": "http://localhost",
                "token": "abcd1234",
            },
            "help": {
                "email": "help-bot@zulip.com",
                "key": "123456789qwertyuiop",
                "site": "http://localhost",
                "token": "abcd1234",
            },
        }
        self.assert_bot_server_response(
            available_bots=available_bots,
            event=dict(
                message={"content": "@**test** test message"},
                bot_email="helloworld-bot@zulip.com",
                trigger="mention",
                token="abcd1234",  # noqa: S106
            ),
            expected_response="beep boop",
            bots_config=bots_config,
            check_success=True,
        )

    def test_request_for_unkown_bot(self) -> None:
        bots_config = {
            "helloworld": {
                "email": "helloworld-bot@zulip.com",
                "key": "123456789qwertyuiop",
                "site": "http://localhost",
                "token": "abcd1234",
            },
        }
        self.assert_bot_server_response(
            available_bots=["helloworld"],
            event=dict(message={"content": "test message"}, bot_email="unknown-bot@zulip.com"),
            bots_config=bots_config,
            check_success=False,
        )

    def test_wrong_bot_token(self) -> None:
        available_bots = ["helloworld"]
        bots_config = {
            "helloworld": {
                "email": "helloworld-bot@zulip.com",
                "key": "123456789qwertyuiop",
                "site": "http://localhost",
                "token": "abcd1234",
            }
        }
        self.assert_bot_server_response(
            available_bots=available_bots,
            bots_config=bots_config,
            event=dict(
                message={"content": "@**test** test message"},
                bot_email="helloworld-bot@zulip.com",
                trigger="mention",
                token="wrongtoken",  # noqa: S106
            ),
            check_success=False,
        )

    @mock.patch("logging.error")
    @mock.patch("zulip_bots.lib.StateHandler")
    def test_wrong_bot_credentials(
        self, mock_state_handler: mock.Mock, mock_logging_error: mock.Mock
    ) -> None:
        available_bots = ["nonexistent-bot"]
        bots_config = {
            "nonexistent-bot": {
                "email": "helloworld-bot@zulip.com",
                "key": "123456789qwertyuiop",
                "site": "http://localhost",
                "token": "abcd1234",
            }
        }
        with self.assertRaisesRegex(
            SystemExit,
            'Error: Bot "nonexistent-bot" doesn\'t exist. Please make '
            "sure you have set up the botserverrc file correctly.",
        ):
            self.assert_bot_server_response(
                available_bots=available_bots,
                event=dict(
                    message={"content": "@**test** test message"},
                    bot_email="helloworld-bot@zulip.com",
                    trigger="mention",
                    token="abcd1234",  # noqa: S106
                ),
                bots_config=bots_config,
            )

    @mock.patch("sys.argv", ["zulip-botserver", "--config-file", "/foo/bar/baz.conf"])
    def test_argument_parsing_defaults(self) -> None:
        opts = parse_args()
        assert opts.config_file == "/foo/bar/baz.conf"
        assert opts.bot_name is None
        assert opts.bot_config_file is None
        assert opts.hostname == "127.0.0.1"
        assert opts.port == 5002

    def test_read_config_from_env_vars(self) -> None:
        # We use an OrderedDict so that the order of the entries in
        # the stringified environment variable is standard even on
        # Python 3.7 and earlier.
        bots_config = OrderedDict()
        bots_config["hello_world"] = {
            "email": "helloworld-bot@zulip.com",
            "key": "value",
            "site": "http://localhost",
            "token": "abcd1234",
        }
        bots_config["giphy"] = {
            "email": "giphy-bot@zulip.com",
            "key": "value2",
            "site": "http://localhost",
            "token": "abcd1234",
        }
        os.environ["ZULIP_BOTSERVER_CONFIG"] = json.dumps(bots_config)

        # No bot specified; should read all bot configs
        assert server.read_config_from_env_vars() == bots_config

        # Specified bot exists; should read only that section.
        assert server.read_config_from_env_vars("giphy") == {"giphy": bots_config["giphy"]}

        # Specified bot doesn't exist; should read the first section of the config.
        assert server.read_config_from_env_vars("redefined_bot") == {
            "redefined_bot": bots_config["hello_world"]
        }

    def test_read_config_file(self) -> None:
        with self.assertRaises(IOError):
            server.read_config_file("nonexistentfile.conf")
        current_dir = os.path.dirname(os.path.abspath(__file__))

        # No bot specified; should read all bot configs.
        bot_conf1 = server.read_config_file(os.path.join(current_dir, "test.conf"))
        expected_config1 = {
            "helloworld": {
                "email": "helloworld-bot@zulip.com",
                "key": "value",
                "site": "http://localhost",
                "token": "abcd1234",
            },
            "giphy": {
                "email": "giphy-bot@zulip.com",
                "key": "value2",
                "site": "http://localhost",
                "token": "abcd1234",
            },
        }
        assert json.dumps(bot_conf1, sort_keys=True) == json.dumps(expected_config1, sort_keys=True)

        # Specified bot exists; should read only that section.
        bot_conf3 = server.read_config_file(os.path.join(current_dir, "test.conf"), "giphy")
        expected_config3 = {
            "giphy": {
                "email": "giphy-bot@zulip.com",
                "key": "value2",
                "site": "http://localhost",
                "token": "abcd1234",
            }
        }
        assert json.dumps(bot_conf3, sort_keys=True) == json.dumps(expected_config3, sort_keys=True)

        # Specified bot doesn't exist; should read the first section of the config.
        bot_conf2 = server.read_config_file(os.path.join(current_dir, "test.conf"), "redefined_bot")
        expected_config2 = {
            "redefined_bot": {
                "email": "helloworld-bot@zulip.com",
                "key": "value",
                "site": "http://localhost",
                "token": "abcd1234",
            }
        }
        assert json.dumps(bot_conf2, sort_keys=True) == json.dumps(expected_config2, sort_keys=True)

    def test_load_lib_modules(self) -> None:
        # This testcase requires hardcoded paths, which here is a good thing so if we ever
        # restructure zulip_bots, this test would fail and we would also update Botserver
        # at the same time.
        helloworld = import_module("zulip_bots.bots.{bot}.{bot}".format(bot="helloworld"))
        root_dir = Path(__file__).parents[2].as_posix()
        # load valid module name
        module = server.load_lib_modules(["helloworld"])["helloworld"]
        assert module == helloworld

        # load valid file path
        path = Path(
            root_dir, "zulip_bots/zulip_bots/bots/{bot}/{bot}.py".format(bot="helloworld")
        ).as_posix()
        module = server.load_lib_modules([path])[path]
        assert module.__name__ == "custom_bot_module"
        assert module.__file__ == path
        assert isinstance(module, ModuleType)

        # load invalid module name
        with self.assertRaisesRegex(
            SystemExit,
            'Error: Bot "botserver-test-case-random-bot" doesn\'t exist. '
            "Please make sure you have set up the botserverrc file correctly.",
        ):
            module = server.load_lib_modules(["botserver-test-case-random-bot"])[
                "botserver-test-case-random-bot"
            ]

        # load invalid file path
        with self.assertRaisesRegex(
            SystemExit,
            f'Error: Bot "{root_dir}/zulip_bots/zulip_bots/bots/helloworld.py" doesn\'t exist. '
            "Please make sure you have set up the botserverrc file correctly.",
        ):
            path = Path(
                root_dir, "zulip_bots/zulip_bots/bots/{bot}.py".format(bot="helloworld")
            ).as_posix()
            module = server.load_lib_modules([path])[path]

    @mock.patch("zulip_botserver.server.app")
    @mock.patch("sys.argv", ["zulip-botserver", "--config-file", "/foo/bar/baz.conf"])
    def test_load_from_registry(self, mock_app: mock.Mock) -> None:
        packaged_bot_module = mock.MagicMock(__version__="1.0.0", __file__="asd")
        packaged_bot_entrypoint = metadata.EntryPoint(
            "packaged_bot", "module_name", "zulip_bots.registry"
        )
        bots_config = {
            "packaged_bot": {
                "email": "packaged-bot@zulip.com",
                "key": "value",
                "site": "http://localhost",
                "token": "abcd1234",
            }
        }

        with mock.patch(
            "zulip_botserver.server.read_config_file", return_value=bots_config
        ), mock.patch("zulip_botserver.server.lib.ExternalBotHandler", new=mock.Mock()), mock.patch(
            "zulip_bots.finder.metadata.EntryPoint.load",
            return_value=packaged_bot_module,
        ), mock.patch(
            "zulip_bots.finder.metadata.entry_points",
            return_value=(packaged_bot_entrypoint,),
        ):
            server.main()

        mock_app.config.__setitem__.assert_any_call(
            "BOTS_LIB_MODULES", {"packaged_bot": packaged_bot_module}
        )



#!/usr/bin/env python3

import configparser
import json
import logging
import os
import sys
from collections import OrderedDict
from configparser import MissingSectionHeaderError, NoOptionError
from importlib import import_module
from types import ModuleType
from typing import Any, Dict, List, Optional

from flask import Flask, request
from werkzeug.exceptions import BadRequest, Unauthorized

from zulip import Client
from zulip_bots import lib
from zulip_bots.finder import import_module_from_source, import_module_from_zulip_bot_registry
from zulip_botserver.input_parameters import parse_args


def read_config_section(parser: configparser.ConfigParser, section: str) -> Dict[str, str]:
    section_info = {
        "email": parser.get(section, "email"),
        "key": parser.get(section, "key"),
        "site": parser.get(section, "site"),
        "token": parser.get(section, "token"),
    }
    return section_info


def read_config_from_env_vars(bot_name: Optional[str] = None) -> Dict[str, Dict[str, str]]:
    bots_config: Dict[str, Dict[str, str]] = {}
    json_config = os.environ.get("ZULIP_BOTSERVER_CONFIG")

    if json_config is None:
        raise OSError(
            "Could not read environment variable 'ZULIP_BOTSERVER_CONFIG': Variable not set."
        )

    # Load JSON-formatted environment variable; use OrderedDict to
    # preserve ordering on Python 3.6 and below.
    env_config = json.loads(json_config, object_pairs_hook=OrderedDict)
    if bot_name is not None:
        if bot_name in env_config:
            bots_config[bot_name] = env_config[bot_name]
        else:
            # If the bot name provided via the command line does not
            # exist in the configuration provided via the environment
            # variable, use the first bot in the environment variable,
            # with name updated to match, along with a warning.
            first_bot_name = next(iter(env_config.keys()))
            bots_config[bot_name] = env_config[first_bot_name]
            logging.warning(
                "First bot name in the config list was changed from %r to %r",
                first_bot_name,
                bot_name,
            )
    else:
        bots_config = dict(env_config)
    return bots_config


def read_config_file(
    config_file_path: str, bot_name: Optional[str] = None
) -> Dict[str, Dict[str, str]]:
    parser = parse_config_file(config_file_path)

    bots_config: Dict[str, Dict[str, str]] = {}
    if bot_name is None:
        bots_config = {
            section: read_config_section(parser, section) for section in parser.sections()
        }
        return bots_config

    logging.warning("Single bot mode is enabled")
    if len(parser.sections()) == 0:
        sys.exit(
            "Error: Your Botserver config file `{0}` does not contain any sections!\n"
            "You need to write the name of the bot you want to run in the "
            "section header of `{0}`.".format(config_file_path)
        )

    if bot_name in parser.sections():
        bot_section = bot_name
        bots_config[bot_name] = read_config_section(parser, bot_name)
        ignored_sections = [section for section in parser.sections() if section != bot_name]
    else:
        bot_section = parser.sections()[0]
        bots_config[bot_name] = read_config_section(parser, bot_section)
        logging.warning(
            "First bot name in the config list was changed from %r to %r", bot_section, bot_name
        )
        ignored_sections = parser.sections()[1:]

    if len(ignored_sections) > 0:
        logging.warning("Sections except the %r will be ignored", bot_section)

    return bots_config


def parse_config_file(config_file_path: str) -> configparser.ConfigParser:
    config_file_path = os.path.abspath(os.path.expanduser(config_file_path))
    if not os.path.isfile(config_file_path):
        raise OSError(f"Could not read config file {config_file_path}: File not found.")
    parser = configparser.ConfigParser()
    parser.read(config_file_path)
    return parser


def load_lib_modules(available_bots: List[str]) -> Dict[str, ModuleType]:
    bots_lib_module = {}
    for bot in available_bots:
        try:
            if bot.endswith(".py") and os.path.isfile(bot):
                lib_module = import_module_from_source(bot, "custom_bot_module")
            else:
                module_name = f"zulip_bots.bots.{bot}.{bot}"
                lib_module = import_module(module_name)
            bots_lib_module[bot] = lib_module
        except ImportError:
            _, bots_lib_module[bot] = import_module_from_zulip_bot_registry(bot)
            if bots_lib_module[bot] is None:
                error_message = (
                    f'Error: Bot "{bot}" doesn\'t exist. Please make sure '
                    "you have set up the botserverrc file correctly.\n"
                )
                if bot == "api":
                    error_message += (
                        "Did you forget to specify the bot you want to run with -b <botname> ?"
                    )
                sys.exit(error_message)
    return bots_lib_module


def load_bot_handlers(
    available_bots: List[str],
    bot_lib_modules: Dict[str, ModuleType],
    bots_config: Dict[str, Dict[str, str]],
    third_party_bot_conf: Optional[configparser.ConfigParser] = None,
) -> Dict[str, lib.ExternalBotHandler]:
    bot_handlers = {}
    for bot in available_bots:
        client = Client(
            email=bots_config[bot]["email"],
            api_key=bots_config[bot]["key"],
            site=bots_config[bot]["site"],
        )
        bot_file = bot_lib_modules[bot].__file__
        assert bot_file is not None
        bot_dir = os.path.dirname(os.path.abspath(bot_file))
        bot_handler = lib.ExternalBotHandler(
            client, bot_dir, bot_details={}, bot_config_parser=third_party_bot_conf
        )

        bot_handlers[bot] = bot_handler
    return bot_handlers


def init_message_handlers(
    available_bots: List[str],
    bots_lib_modules: Dict[str, Any],
    bot_handlers: Dict[str, lib.ExternalBotHandler],
) -> Dict[str, Any]:
    message_handlers = {}
    for bot in available_bots:
        bot_lib_module = bots_lib_modules[bot]
        bot_handler = bot_handlers[bot]
        message_handler = lib.prepare_message_handler(bot, bot_handler, bot_lib_module)
        message_handlers[bot] = message_handler
    return message_handlers


app = Flask(__name__)
bots_config: Dict[str, Dict[str, str]] = {}


@app.route("/", methods=["POST"])
def handle_bot() -> str:
    event = request.get_json(force=True)
    assert event is not None
    for bot_name, config in bots_config.items():
        if config["email"] == event["bot_email"]:
            bot = bot_name
            bot_config = config
            break
    else:
        raise BadRequest(
            "Cannot find a bot with email {} in the Botserver "
            "configuration file. Do the emails in your botserverrc "
            "match the bot emails on the server?".format(event["bot_email"])
        )
    if bot_config["token"] != event["token"]:
        raise Unauthorized(
            "Request token does not match token found for bot {} in the "
            "Botserver configuration file. Do the outgoing webhooks in "
            "Zulip point to the right Botserver?".format(event["bot_email"])
        )
    app.config.get("BOTS_LIB_MODULES", {})[bot]
    bot_handler = app.config.get("BOT_HANDLERS", {})[bot]
    message_handler = app.config.get("MESSAGE_HANDLERS", {})[bot]
    is_mentioned = event["trigger"] == "mention"
    # TODO/compatibility: Remove the support for "private_message" as a valid
    # trigger value once we no longer support pre-8.0 Zulip servers.
    is_direct_message = event["trigger"] in ["direct_message", "private_message"]
    message = event["message"]
    message["full_content"] = message["content"]
    # Strip at-mention botname from the message
    if is_mentioned:
        # message['content'] will be None when the bot's @-mention is not at the beginning.
        # In that case, the message shall not be handled.
        message["content"] = lib.extract_query_without_mention(message=message, client=bot_handler)
        if message["content"] is None:
            return json.dumps(dict(response_not_required=True))

    if is_direct_message or is_mentioned:
        message_handler.handle_message(message=message, bot_handler=bot_handler)
    return json.dumps(dict(response_not_required=True))


def main() -> None:
    options = parse_args()
    global bots_config  # noqa: PLW0603

    if options.use_env_vars:
        bots_config = read_config_from_env_vars(options.bot_name)
    elif options.config_file:
        try:
            bots_config = read_config_file(options.config_file, options.bot_name)
        except MissingSectionHeaderError:
            sys.exit(
                "Error: Your Botserver config file `{0}` contains an empty section header!\n"
                "You need to write the names of the bots you want to run in the "
                "section headers of `{0}`.".format(options.config_file)
            )
        except NoOptionError as e:
            sys.exit(
                "Error: Your Botserver config file `{0}` has a missing option `{1}` in section `{2}`!\n"
                "You need to add option `{1}` with appropriate value in section `{2}` of `{0}`".format(
                    options.config_file, e.option, e.section
                )
            )

    available_bots = list(bots_config.keys())
    bots_lib_modules = load_lib_modules(available_bots)
    third_party_bot_conf = (
        parse_config_file(options.bot_config_file) if options.bot_config_file is not None else None
    )
    bot_handlers = load_bot_handlers(
        available_bots, bots_lib_modules, bots_config, third_party_bot_conf
    )
    message_handlers = init_message_handlers(available_bots, bots_lib_modules, bot_handlers)
    app.config["BOTS_LIB_MODULES"] = bots_lib_modules
    app.config["BOT_HANDLERS"] = bot_handlers
    app.config["MESSAGE_HANDLERS"] = message_handlers
    app.run(host=options.hostname, port=int(options.port))


if __name__ == "__main__":
    main()



from setuptools import find_packages, setup

ZULIP_BOTSERVER_VERSION = "0.9.1"

with open("README.md") as fh:
    long_description = fh.read()

setup(
    name="zulip_botserver",
    version=ZULIP_BOTSERVER_VERSION,
    description="Zulip's Flask server for running bots",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Zulip Open Source Project",
    author_email="zulip-devel@googlegroups.com",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Communications :: Chat",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    python_requires=">=3.9",
    url="https://www.zulip.org/",
    project_urls={
        "Source": "https://github.com/zulip/python-zulip-api/",
        "Documentation": "https://zulip.com/api",
    },
    entry_points={
        "console_scripts": [
            "zulip-botserver=zulip_botserver.server:main",
        ],
    },
    test_suite="tests",
    package_data={"zulip_botserver": ["py.typed"]},
    install_requires=[
        "zulip",
        "zulip_bots",
        "flask>=0.12.2",
        "typing_extensions>=4.5.0",
    ],
    packages=find_packages(exclude=["tests"]),
)
